"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.RecommendationController = void 0;
const errors_1 = __importDefault(require("@tryghost/errors"));
function validateString(object, key, { required = true } = {}) {
    if (typeof object !== 'object' || object === null) {
        throw new errors_1.default.BadRequestError({ message: `${key} must be an object` });
    }
    if (object[key] !== undefined && object[key] !== null) {
        if (typeof object[key] !== "string") {
            throw new errors_1.default.BadRequestError({ message: `${key} must be a string` });
        }
        return object[key];
    }
    else if (required) {
        throw new errors_1.default.BadRequestError({ message: `${key} is required` });
    }
}
function validateBoolean(object, key, { required = true } = {}) {
    if (typeof object !== 'object' || object === null) {
        throw new errors_1.default.BadRequestError({ message: `${key} must be an object` });
    }
    if (object[key] !== undefined) {
        if (typeof object[key] !== "boolean") {
            throw new errors_1.default.BadRequestError({ message: `${key} must be a boolean` });
        }
        return object[key];
    }
    else if (required) {
        throw new errors_1.default.BadRequestError({ message: `${key} is required` });
    }
}
function validateURL(object, key, { required = true } = {}) {
    const string = validateString(object, key, { required });
    if (string !== undefined) {
        try {
            return new URL(string);
        }
        catch (e) {
            throw new errors_1.default.BadRequestError({ message: `${key} must be a valid URL` });
        }
    }
}
class RecommendationController {
    service;
    constructor(deps) {
        this.service = deps.service;
    }
    #getFrameId(frame) {
        if (!frame.options) {
            throw new errors_1.default.BadRequestError();
        }
        const id = frame.options.id;
        if (!id) {
            throw new errors_1.default.BadRequestError();
        }
        return id;
    }
    #getFrameRecommendation(frame) {
        if (!frame.data || !frame.data.recommendations || !frame.data.recommendations[0]) {
            throw new errors_1.default.BadRequestError();
        }
        const recommendation = frame.data.recommendations[0];
        const cleanedRecommendation = {
            title: validateString(recommendation, "title") ?? '',
            url: validateURL(recommendation, "url"),
            // Optional fields
            oneClickSubscribe: validateBoolean(recommendation, "one_click_subscribe", { required: false }) ?? false,
            reason: validateString(recommendation, "reason", { required: false }) ?? null,
            excerpt: validateString(recommendation, "excerpt", { required: false }) ?? null,
            featuredImage: validateURL(recommendation, "featured_image", { required: false }) ?? null,
            favicon: validateURL(recommendation, "favicon", { required: false }) ?? null,
        };
        // Create a new recommendation
        return cleanedRecommendation;
    }
    #getFrameRecommendationEdit(frame) {
        if (!frame.data || !frame.data.recommendations || !frame.data.recommendations[0]) {
            throw new errors_1.default.BadRequestError();
        }
        const recommendation = frame.data.recommendations[0];
        const cleanedRecommendation = {
            title: validateString(recommendation, "title", { required: false }),
            url: validateURL(recommendation, "url", { required: false }),
            oneClickSubscribe: validateBoolean(recommendation, "one_click_subscribe", { required: false }),
            reason: validateString(recommendation, "reason", { required: false }),
            excerpt: validateString(recommendation, "excerpt", { required: false }),
            featuredImage: validateURL(recommendation, "featured_image", { required: false }),
            favicon: validateURL(recommendation, "favicon", { required: false }),
        };
        // Create a new recommendation
        return cleanedRecommendation;
    }
    #returnRecommendations(...recommendations) {
        return {
            data: recommendations.map(r => {
                return {
                    id: r.id,
                    title: r.title,
                    reason: r.reason,
                    excerpt: r.excerpt,
                    featured_image: r.featuredImage?.toString() ?? null,
                    favicon: r.favicon?.toString() ?? null,
                    url: r.url.toString(),
                    one_click_subscribe: r.oneClickSubscribe,
                    created_at: r.createdAt,
                    updated_at: r.updatedAt
                };
            })
        };
    }
    async addRecommendation(frame) {
        const recommendation = this.#getFrameRecommendation(frame);
        return this.#returnRecommendations(await this.service.addRecommendation(recommendation));
    }
    async editRecommendation(frame) {
        const id = this.#getFrameId(frame);
        const recommendationEdit = this.#getFrameRecommendationEdit(frame);
        return this.#returnRecommendations(await this.service.editRecommendation(id, recommendationEdit));
    }
    async deleteRecommendation(frame) {
        const id = this.#getFrameId(frame);
        await this.service.deleteRecommendation(id);
    }
    async listRecommendations() {
        return this.#returnRecommendations(...(await this.service.listRecommendations()));
    }
}
exports.RecommendationController = RecommendationController;
//# sourceMappingURL=RecommendationController.js.map